#
#  mx_guifn.tcl -- Functions used by the MX GUI.
#
#  Written by William Lavender.
#
#------------------------------------------------------------------------
#
#  Copyright 1999-2004 Illinois Institute of Technology
#
#  See the file "LICENSE" for information on usage and redistribution
#  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#

package provide MxGui 1.0

namespace eval MxGui {
	variable measurement_time
	variable motor_polling_interval
	variable overwrite_enabled
	variable monospaced_font
	variable textbackground
	variable user_interrupt
}

#
#  Setup the MX gui.
#

proc setup_mxgui { argv0 argv } {

	global errorInfo

	wm title . "MX Gui"
	
	#
	#  Initialize the MX Tcl library.
	#
	
	set status [ catch { ::Mx::init } result ]

	if { $status != 0 } {
		Mx::error_message "MX Initialization Failed" \
		"The MX initialization procedure Mx::init failed.  Exiting..." \
			$result $errorInfo
		exit
	}

	set MxGui::user_interrupt 0
	
	#
	#  Now setup the GUI and the database.
	#

	set textbackground white

	set MxGui::textbackground $textbackground

	option add *Entryfield.textBackground $textbackground
	option add *Promptdialog.textBackground $textbackground
	option add *Selectiondialog.textBackground $textbackground
	option add *Scrolledtext.textBackground $textbackground
	
	set MxGui::monospaced_font \
		[ font create -family Courier ]

	set MxGui::measurement_time 1

	set MxGui::motor_polling_interval 500

	set MxGui::overwrite_enabled 0

	set status [ catch { create_message_log_window } result ]

	if { $status != 0 } {
		Mx::error_message "Can't Create Message Log" \
	"The attempt to create a message log window failed.  Exiting..." \
			$result $errorInfo
		exit
	}
	
	log_puts "*** Initializing MX GUI.  Please wait. ***"
	
	set version [ Mx::get_version ]
	
	log_puts ""
	log_puts "Using MX version $version"
	log_puts ""

	set status [ catch { setup_mx_database $argv } result ]

	if { $status != 0 } {
		Mx::error_message "MX Initialization Failed" \
	"Unable to process one or more of the MX database files.  Exiting..." \
		    $result $errorInfo
	
		exit
	}

	set record_list $result

	set status [ catch { create_menu_bar $record_list } result ]

	if { $status != 0 } {
		Mx::error_message "Cannot Create Menu Bar" \
		"The attempt to create the menu bar failed.  Exiting..." \
			$result $errorInfo
		exit
	}

	set argv0 [ file tail $argv0 ]

	set status [ catch { $record_list set_program_name $argv0 } ]

	if { $status != 0 } {
		log_puts \
		"Unable to set the MX program name to 'mxgui' for all servers."
	}
	
	log_puts "*** MX GUI initialization complete. ***"
	
}

#
#  Setup the MX record database.
#
proc setup_mx_database { argv } {

	set argv_list [ split $argv ]

	set motor_database [ lindex $argv_list 0 ]

	set argv_list [ lrange $argv_list 1 end ]

	#
	#  Create a new record list.
	#

	set record_list [ Mx::RecordList ::mxrecordlist_#auto ]

	#
	#  Enable plotting.
	#

	$record_list set_plot_enable 1

	#
	#  Read the motor database.
	#

	$record_list read_database $motor_database

	$record_list finish_database_initialization

	update

	#
	#  Read all of the scan databases.
	#

	foreach file $argv_list {

		if [ file readable $file ] {
			$record_list load_scans $file 1
		} else {
			if [ file exists $file ] {
				log_puts \
	"Scan database '$file' exists, but is not readable.  Skipping it."

			} else {
				log_puts \
			"Scan database '$file' does not exist.  Skipping it."
			}
		}
		update
	}

	#
	#  Finish initializing the MX system.
	#

	$record_list initialize_hardware 0

	$record_list configure -unsaved_scans 0

	#
	#  Install an event handler to allow Tk events to be processed
	#  while MX is scanning.
	#

	Mx::set_user_interrupt_function mxgui_user_interrupt_function

	#
	#  Setup the log handler.
	#

	$record_list log_open

	return $record_list
}

proc mxgui_user_interrupt_function {} {

	update

	set interrupt_occurred $MxGui::user_interrupt

	set MxGui::user_interrupt 0

	return $interrupt_occurred
}

#
#  Create a message log window.
#

proc create_message_log_window {} {

	set window .message_log

	#
	#  Create the window.
	#

	iwidgets::scrolledtext $window

	$window configure -visibleitems 80x24
	$window configure -hscrollmode dynamic
	$window configure -vscrollmode dynamic

	pack $window -expand yes -fill both

	#
	#  Install handlers to redirect the output of mx_info(), mx_error()
	#  and MX_DEBUG() to this window.
	#

	Mx::set_all_output_functions mxgui_output_function

	#
	#  Install handler for mx_info_dialog().
	#

	Mx::set_info_dialog_function mxgui_info_dialog_function

}

proc mxgui_output_function { message } {
	log_puts $message
	update
}

proc mxgui_info_dialog_function { text_prompt gui_prompt button_label } {
	bell
	tk_dialog .mxgui_info_dialog "MX Info" $gui_prompt info 0 $button_label
}

##########

proc log_puts { message } {
	set textwindow [ .message_log childsite ]

	$textwindow configure -state normal

	.message_log insert end "$message\n"

	$textwindow configure -state disabled

	.message_log yview moveto 1.0
}

# Datafile descriptions and plot descriptions have optional type arguments
# that are separated from the type by a colon ':' character.  For now,
# we discard any optional type arguments.  (WML)

proc get_datafile_type { scan_record } {

	set description [ $scan_record get_field datafile_description ]

	set list [ split $description : ]

	set result [ lrange $list 0 0 ]

	return $result
}

proc get_plot_type { scan_record } {

	set description [ $scan_record get_field plot_description ]

	set list [ split $description : ]

	set result [ lrange $list 0 0 ]

	return $result
}

#
#  Create the menu bar for the top level menu.
#

proc create_menu_bar { record_list } {
	global env

	menu .menubar
	. configure -menu .menubar

	###################################################################
	#
	#  Add the File menu.
	#
	menu .menubar.file -tearoff 0
	.menubar add cascade -label File -menu .menubar.file

	.menubar.file add command -label Exit -command \
				[ itcl::code menubar_exit $record_list ]

	#------------------------------------------------------------------

	proc menubar_exit { record_list } {

		#
		#  Shut down the log handler.
		#

		$record_list log_close

		#
		#  Save any unsaved scans.
		#

		set unsaved_scans [ $record_list cget -unsaved_scans ]

		if { $unsaved_scans != 0 } {
			set t \
	"You have unsaved scans.\nDo you want to save them before exiting?"

			set selection [tk_dialog .motor_unsaved_scans_dialog \
			    "Unsaved Scans" $t question 0 "Yes" "No" "Cancel"]

			if { $selection == 0 } {
				center_on_window . .scan_load_save_dialog

				set result [ .scan_load_save_dialog activate ]

				if { $result == 0 } {
					# We got here by hitting the Cancel
					# button, so return to the main menu.

					return
				}

				set filename [ .scan_load_save_dialog get ]

				$record_list save_scans $filename

				exit
			}
			if { $selection == 1 } {
				exit
			}

			# Otherwise just return to the scan setup screen.

			return
		}
		exit
	}

	###################################################################
	#
	#  Add the Edit menu.
	#
	menu .menubar.edit -tearoff 0
	.menubar add cascade -label Edit -menu .menubar.edit

	.menubar.edit add command -label "Set Debug Level" \
				-command menubar_set_debug_level_dialog
	.menubar.edit add command -label "Set Motor Polling Interval" \
				-command menubar_set_polling_interval_dialog
	.menubar.edit add command -label "Set Plot Enable Flag" \
	    -command [ itcl::code menubar_set_plot_enable_dialog $record_list ]
	.menubar.edit add command -label "Set Overwrite Enable Flag" \
    -command [ itcl::code menubar_set_overwrite_enable_dialog $record_list ]

	proc menubar_set_debug_level_dialog {} {
		iwidgets::promptdialog .set_debug_level_dialog

		.set_debug_level_dialog configure -title "Set Debug Level"
		.set_debug_level_dialog configure -labeltext "Debug Level"
		.set_debug_level_dialog hide Apply
		.set_debug_level_dialog configure -validate \
					{ iwidgets::Entryfield::integer %P }

		.set_debug_level_dialog buttonconfigure OK -command \
						menubar_set_debug_level_ok
		.set_debug_level_dialog buttonconfigure Cancel -command \
						menubar_set_debug_level_cancel

		set debug_level [ Mx::get_debug_level ]

		.set_debug_level_dialog insert 0 [ Mx::get_debug_level ]

		center_on_window . .set_debug_level_dialog

		.set_debug_level_dialog activate
	}

	proc menubar_set_debug_level_ok {} {
		set debug_level [ .set_debug_level_dialog get ]

		.set_debug_level_dialog deactivate
		destroy .set_debug_level_dialog

		Mx::set_debug_level $debug_level
	}

	proc menubar_set_debug_level_cancel {} {
		.set_debug_level_dialog deactivate
		destroy .set_debug_level_dialog
	}

	###################################################################

	proc menubar_set_polling_interval_dialog {} {
		iwidgets::promptdialog .polling_dialog

		.polling_dialog configure -title "Set Motor Polling Interval"
		.polling_dialog configure \
			-labeltext "Motor Polling Interval (milliseconds)"
		.polling_dialog hide Apply
		.polling_dialog configure -validate \
					{ iwidgets::Entryfield::integer %P }

		.polling_dialog buttonconfigure OK -command \
					menubar_set_polling_interval_ok
		.polling_dialog buttonconfigure Cancel -command \
					menubar_set_polling_interval_cancel

		.polling_dialog insert 0 $MxGui::motor_polling_interval

		center_on_window . .polling_dialog

		.polling_dialog activate
	}

	proc menubar_set_polling_interval_ok {} {
		set MxGui::motor_polling_interval [ .polling_dialog get ]

		.polling_dialog deactivate
		destroy .polling_dialog
	}

	proc menubar_set_polling_interval_cancel {} {
		.polling_dialog deactivate
		destroy .polling_dialog
	}

	###################################################################

	proc menubar_set_plot_enable_dialog { record_list } {
		set option_list [ list "Off" "On" "No Wait" ]

		set dialog .set_plot_enable_dialog

		toplevel $dialog

		wm withdraw $dialog

		wm title $dialog "Set Plot Enable Flag"

		#------------------------------------------------

		iwidgets::optionmenu $dialog.option
		pack $dialog.option

		$dialog.option configure -labeltext "Plot Enable Flag"

		foreach entry $option_list {
			$dialog.option insert end $entry
		}

		set plot_enable_flag [ $record_list plotting_is_enabled ]

		$dialog.option select [ lindex $option_list $plot_enable_flag ]

		#------------------------------------------------

		frame $dialog.sep -height 2 -borderwidth 1 -relief sunken
		pack $dialog.sep -expand yes -fill both

		#------------------------------------------------

		iwidgets::buttonbox $dialog.buttonbox
		pack $dialog.buttonbox -expand yes -fill both

		$dialog.buttonbox add Accept -text Accept -command \
	[ itcl::code menubar_set_plot_enable_ok $record_list $option_list ]

		$dialog.buttonbox add Cancel -text Cancel -command \
					{ destroy .set_plot_enable_dialog }

		$dialog.buttonbox default Accept

		#------------------------------------------------

		center_on_window . $dialog

		wm deiconify $dialog
	}

	proc menubar_set_plot_enable_ok { record_list option_list } {

		set plot_enable_label [ .set_plot_enable_dialog.option get ]

		set plot_enable_flag \
			[ lsearch -exact $option_list "$plot_enable_label" ]

		destroy .set_plot_enable_dialog

		$record_list set_plot_enable $plot_enable_flag
	}

	###################################################################

	proc menubar_set_overwrite_enable_dialog { record_list } {
		set option_list [ list "False" "True" ]

		set dialog .set_overwrite_enable_dialog

		toplevel $dialog

		wm withdraw $dialog

		wm title $dialog "Set Overwrite Enable Flag"

		#------------------------------------------------

		iwidgets::optionmenu $dialog.option
		pack $dialog.option

		$dialog.option configure -labeltext "Overwrite Enable Flag"

		foreach entry $option_list {
			$dialog.option insert end $entry
		}

		$dialog.option select \
			[ lindex $option_list $MxGui::overwrite_enabled ]

		#------------------------------------------------

		frame $dialog.sep -height 2 -borderwidth 1 -relief sunken
		pack $dialog.sep -expand yes -fill both

		#------------------------------------------------

		iwidgets::buttonbox $dialog.buttonbox
		pack $dialog.buttonbox -expand yes -fill both

		$dialog.buttonbox add Accept -text Accept -command \
	[ itcl::code menubar_set_overwrite_enable_ok $record_list $option_list ]

		$dialog.buttonbox add Cancel -text Cancel -command \
					{ destroy .set_overwrite_enable_dialog }

		$dialog.buttonbox default Accept

		#------------------------------------------------

		center_on_window . $dialog

		wm deiconify $dialog
	}

	proc menubar_set_overwrite_enable_ok { record_list option_list } {

		set overwrite_enable_label \
			[ .set_overwrite_enable_dialog.option get ]

		set MxGui::overwrite_enabled \
		    [ lsearch -exact $option_list "$overwrite_enable_label" ]

		destroy .set_overwrite_enable_dialog
	}

	###################################################################
	#
	#  Add the Interface menu.
	#
#	menu .menubar.interface -tearoff 0
#	.menubar add cascade -label Interface -menu .menubar.interface \
#			-state disabled

	###################################################################
	#
	#  Add the Device menu.
	#
	menu .menubar.device -tearoff 0
	.menubar add cascade -label Device -menu .menubar.device

	.menubar.device add command -label "Move Motor" \
		-command [ itcl::code menubar_move_motor_command $record_list ]
	.menubar.device add command -label "Show Motors" \
		-command [ itcl::code menubar_show_motors_command $record_list ]
	.menubar.device add command -label "Show Scalers" \
		-command [ itcl::code show_scalers_command $record_list ]

	#------------------------------------------------------------------

	iwidgets::selectiondialog .motor_select_dialog

	.motor_select_dialog configure -title "Select motor"
	.motor_select_dialog configure -itemslabel Motors
	.motor_select_dialog configure -textfont $MxGui::monospaced_font

	.motor_select_dialog hide Apply

	.motor_select_dialog buttonconfigure OK -text "Select Motor"
	.motor_select_dialog buttonconfigure OK \
	    -command [ itcl::code menubar_select_motor_ok_command $record_list ]

	set motorlist \
		[ $record_list get_multiple_records "device" "motor" ".*" 0 ]

	foreach motor $motorlist {
		.motor_select_dialog insert items end $motor
	}

	#------------------------------------------------------------------

	proc menubar_move_motor_command { record_list } {
		if { [ .motor_select_dialog size ] <= 0 } {

			Mx::error_message "No Motors Defined" \
	"Error: No motor records are defined for the current database." "" ""
		} else {
			center_on_window . .motor_select_dialog
			.motor_select_dialog activate
		}
	}

	proc menubar_select_motor_ok_command { record_list } {
		.motor_select_dialog deactivate

		set motor_item_string [ .motor_select_dialog get ]

		scan $motor_item_string "%s" motor_name

		set motor_record [ $record_list get_record $motor_name ]

		mxgui_create_motor_widget $motor_record
	}

	#------------------------------------------------------------------

	proc menubar_show_motors_command { record_list } {

		# Create the show motors window.
		
		set mainwindow .show_motors_window

		set status [ catch { toplevel $mainwindow } result ] 

		# If the window already exists, bring it to the top.

		if { $status != 0 } {
			raise $mainwindow
			focus $mainwindow
			return
		}

		wm title $mainwindow "Motor Positions"

		label $mainwindow.status -fg red \
				-text "Not Yet Initialized"

		pack $mainwindow.status

		iwidgets::scrolledframe $mainwindow.motorframe \
			-hscrollmode none -vscrollmode dynamic \
			-height 400 -width 300

		pack $mainwindow.motorframe -expand yes -fill both

		set window [ $mainwindow.motorframe childsite ]

		set row 0

		label $window.motor_name -text "Motor Name" -width 20 
		label $window.position -text "Position" -width 12 
		label $window.units -text "Units" -width 10 

		grid $window.motor_name $window.position $window.units \
			-row $row

		incr row

		set motorlist \
		  [ $record_list get_multiple_records "device" "motor" ".*" 0 ]

		set num_motors [ llength $motorlist ]

		# Don't make the field $mainwindow.num_motors visible.  It is
		# just intended as a place to store the number of motors.

		label $mainwindow.num_motors -text $num_motors

		for { set i 0 } { $i < $num_motors } { incr i } {

			set thismotor $window.motor$i

			set motorname [ lindex $motorlist $i ]

			set thisrecord [ $record_list get_record $motorname ]

			label ${thismotor}_name -text $motorname \
					-anchor w \
					-font $MxGui::monospaced_font \
					-bg $MxGui::textbackground

			label ${thismotor}_position \
					-anchor w \
					-font $MxGui::monospaced_font \
					-bg $MxGui::textbackground

			label ${thismotor}_units -text \
				[ $thisrecord get_field units ] \
					-anchor w \
					-font $MxGui::monospaced_font \
					-bg $MxGui::textbackground

			grid ${thismotor}_name ${thismotor}_position \
				${thismotor}_units -sticky news -row $row

			incr row
		}

		frame $mainwindow.sep -height 2 -borderwidth 1 -relief sunken
		pack $mainwindow.sep

		set buttonbox $mainwindow.buttonbox

		iwidgets::buttonbox $buttonbox

		$buttonbox add Update -text Update
		$buttonbox add Dismiss -text Dismiss
		$buttonbox default Update

		$buttonbox buttonconfigure Update -command [ itcl::code \
			menubar_update_show_motors $record_list $mainwindow ]

		$buttonbox buttonconfigure Dismiss \
				-command [ itcl::code destroy $mainwindow ]

		pack $buttonbox

		menubar_update_show_motors $record_list $mainwindow
	}

	proc menubar_update_show_motors { record_list mainwindow } {

		$mainwindow.status configure -fg yellow
		$mainwindow.status configure -text "Reading Motor Positions"

		update

		set window [ $mainwindow.motorframe childsite ]

		set num_motors [ $mainwindow.num_motors cget -text ]

		for { set i 0 } { $i < $num_motors } { incr i } {

			set thismotor $window.motor$i

			set motorname [ ${thismotor}_name cget -text ]

			set thisrecord [ $record_list get_record $motorname ]

			set position [ format "%.8g" \
					[ $thisrecord get_position ] ]

			${thismotor}_position configure -text $position
		}

		$mainwindow.status configure -fg limegreen
		$mainwindow.status configure -text "Done"

		update idletasks

		set framewidth [ winfo width $window ]

		$mainwindow.motorframe configure -width $framewidth
	}

	###################################################################
	#
	#  Add the Scan menu.
	#
	menu .menubar.scan -tearoff 0
	.menubar add cascade -label Scan -menu .menubar.scan

	.menubar.scan add command -label "New Scan" -command [ itcl::code \
		menubar_new_scan_dialog $record_list ]

	.menubar.scan add command -label "Edit Scan" -command [ itcl::code \
		menubar_scan_dialog $record_list menubar_edit_scan_command ]

	.menubar.scan add command -label "Copy Scan" -command [ itcl::code \
		menubar_scan_dialog $record_list menubar_copy_scan_command ]

	.menubar.scan add command -label "Delete Scan" -command [ itcl::code \
		menubar_scan_dialog $record_list menubar_delete_scan_command ]

	.menubar.scan add command -label "Load Scans" -command [ itcl::code \
		menubar_load_scans_dialog $record_list ]

	.menubar.scan add command -label "Save Scans" -command [ itcl::code \
		menubar_save_scans_dialog $record_list ]

	.menubar.scan add command -label "Run Scan" -command [ itcl::code \
		menubar_scan_dialog $record_list menubar_run_scan_command ]

	#------------------------------------------------------------------

	iwidgets::selectiondialog .scan_select_dialog

	.scan_select_dialog configure -title "Select scan"
	.scan_select_dialog configure -itemslabel Scans
	.scan_select_dialog configure -textfont $MxGui::monospaced_font
	.scan_select_dialog configure -height 400
	.scan_select_dialog configure -width  500

	.scan_select_dialog buttonconfigure OK -text "Select Scan"
	.scan_select_dialog hide Apply

	#------------------------------------------------------------------

	iwidgets::fileselectiondialog .scan_load_save_dialog

	.scan_load_save_dialog configure -modality application
	.scan_load_save_dialog configure -mask "*.dat"

	# Set the starting directory for scans to be loaded from and saved to.

	if [ catch { set mx_tcl_savefile_dir $env(MX_TCL_SAVEFILE_DIR) } ] {

		set mx_tcl_savefile_dir [ pwd ]
	}

	# Does the requested directory exist?

	if { [ file isdirectory $mx_tcl_savefile_dir ] == 0 } {

		set mx_tcl_savefile_dir [ pwd ]
	}

	.scan_load_save_dialog configure -directory $mx_tcl_savefile_dir

	set scan_load_save_default_filename [ file join $mx_tcl_savefile_dir "scan.dat" ]

	set scan_load_save_entryfield \
		.scan_load_save_dialog.shellchildsite.dschildsite.fsb.selection

	$scan_load_save_entryfield insert end $scan_load_save_default_filename

	#------------------------------------------------------------------

	proc menubar_scan_dialog { record_list command } {
		menubar_update_scan_list $record_list

		if { [ .scan_select_dialog size ] <= 0 } {
			Mx::error_message "No Scans Available" \
	"No scan records have been created yet for the current database." "" ""
		} else {
			center_on_window . .scan_select_dialog

			.scan_select_dialog buttonconfigure OK -command \
				[ itcl::code $command $record_list ]

			.scan_select_dialog activate
		}
	}

	proc menubar_update_scan_list { record_list } {
		set scanlist \
			[ $record_list get_multiple_records "scan" ".*" ".*" 1 ]

		.scan_select_dialog clear items

		foreach name $scanlist {
			set scan_record  [ $record_list get_record $name ]
			set scan_type    [ $scan_record get_field mx_type ]
			set motor_string [ $scan_record get_field motor_array ]

			set item_string [ format "%-24s  %-12s  %s" \
					$name $scan_type $motor_string ]

			.scan_select_dialog insert items end $item_string
		}
#		.scan_select_dialog.sb.items sort ascending

		.scan_select_dialog selection set 0
	}

	proc menubar_get_selected_scan { record_list } {
		.scan_select_dialog deactivate

		set scan_item_string [ .scan_select_dialog get ]

		#
		# Don't be confused by the fact that Tcl's equivalent
		# to C's scanf() is called "scan".
		#

		scan $scan_item_string "%s" scan_name

		set scan_record [ $record_list get_record $scan_name ]

		return $scan_record
	}

	#------------------------------------------------------------------

	proc menubar_new_scan_dialog { record_list } {
		set window .new_scan_dialog

		set status [ catch { toplevel $window } result ] 

		# If the window already exists, bring it to the top.

		if { $status != 0 } {
			raise $window
			focus $window
			return
		}

	# Hide the window for now.

		wm withdraw $window
		wm title $window "New Scan"

		set main $window.main

		frame $main

	# Create the scan name line.

		label $main.scan_name_label -text "New Scan Name"

		iwidgets::entryfield $main.scan_name_value -command \
				[ itcl::code menubar_new_scan_ok $record_list ]

		grid $main.scan_name_label $main.scan_name_value \
						-padx 2 -pady 5

	# Create the scan type line.

		label $main.scan_type_label -text "New Scan Type"

		iwidgets::optionmenu $main.scan_type_value 

#		$main.scan_type_value insert end \
#			Motor Input "Theta-2 Theta" Slit List XAFS

		$main.scan_type_value insert end Motor Input Slit XAFS

		grid $main.scan_type_label $main.scan_type_value \
						-padx 2 -pady 2

		pack $main

	# Create a separator

		frame $window.sep -height 2 -borderwidth 1 -relief sunken
		pack $window.sep -expand yes -fill x

	# Create the OK-cancel buttons.

		frame $window.b
		pack $window.b

		set buttonbox $window.b.buttonbox

		iwidgets::buttonbox $buttonbox -padx 0 -pady 2

		$buttonbox add OK -text OK -command \
				[ itcl::code menubar_new_scan_ok $record_list ]

		$buttonbox add Cancel -text Cancel \
					-command [ itcl::code destroy $window ]
		$buttonbox default OK

		pack $buttonbox

	# Make the window appear on the screen.

		wm deiconify $window
	}

	proc menubar_new_scan_ok { record_list } {
		set window .new_scan_dialog

		wm withdraw $window

		set scan_name [ $window.main.scan_name_value get ]
		set scan_type [ $window.main.scan_type_value get ]

		set scan_name [ string trim $scan_name ]

		destroy $window

		set length [ string length $scan_name ]

		if { [expr $length <= 0] } {
			Mx::error_message "Missing Scan Name" \
				"No scan name was specified." "" ""
			return
		}

		set scan_type [ string tolower $scan_type ]

		set result [ string compare scan_type "theta-2 theta" ]

		if { $result == 0 } {
			set scan_type "2theta"
		} else {
			set mytemp "_scan"
			append scan_type $mytemp
		}

		mxgui_scan_setup $record_list $scan_name $scan_type
	}

	#------------------------------------------------------------------

	proc menubar_edit_scan_command { record_list } {
		set scan_record [ menubar_get_selected_scan $record_list ]

		set scan_name [ $scan_record cget -name ]

		set scan_type [ $scan_record get_field mx_type ]

		mxgui_scan_setup $record_list $scan_name $scan_type
	}

	#------------------------------------------------------------------

	proc menubar_copy_scan_command { record_list } {
		set scan_record [ menubar_get_selected_scan $record_list ]

		set scan_name [ $scan_record cget -name ]

		# Get the name of the new scan here.

		iwidgets::promptdialog .copy_scan_new_name

		.copy_scan_new_name configure -title "New scan name"
		.copy_scan_new_name configure -labeltext "New scan name"
		.copy_scan_new_name hide Apply

		.copy_scan_new_name buttonconfigure OK -command [ itcl::code \
			menubar_copy_scan_new_name $record_list $scan_record ]

		center_on_window . .copy_scan_new_name
		.copy_scan_new_name activate
	}

	proc menubar_copy_scan_new_name { record_list old_scan_record } {
		set new_scan_name [ .copy_scan_new_name get ]

		.copy_scan_new_name deactivate
		destroy .copy_scan_new_name

		# Get the old record description and replace the old name
		# in the description with the new name.

		set old_description \
			    [ $old_scan_record create_description_from_record ]

		set old_scan_name [ $old_scan_record cget -name ]

		set old_name_length [ string length $old_scan_name ]
		set new_name_length [ string length $new_scan_name ]

		if { [expr $old_name_length > $new_name_length] } {
			set name_length $old_name_length
		} else {
			set name_length $new_name_length
		}

		set description_tail \
			[ string range $old_description $name_length end ]

		set description $new_scan_name
		append description $description_tail

		# Find out whether the record already exists.

		set new_scan_record [ Mx::object_name $new_scan_name ]

		set result [itcl::find objects $new_scan_record]

		set result_length [llength $result]

		# If the new scan name already exists, check to see
		# whether it is a scan record.

		if { $result_length != 0 } {
			set isa_scan [ $new_scan_record isa Mx::Scan ]

			if { $isa_scan == 0 } {
				set t \
			"A record named '$new_scan_name' already exists,\n"
				append t \
			"but it is not a scan record, so it cannot\n"
				append t \
			"be overwritten."

				Mx::error_message "Illegal Scan Name" \
						$t "" ""
				return
			}

			# The old record is a scan record, so delete it
			# to make way for the new scan record.

			$new_scan_record delete_record
		}

		# Create the new record

		set new_scan_record [ $record_list \
				create_record_from_description $description ]

		$new_scan_record finish_record_initialization

		$record_list configure -unsaved_scans 1
	}

	#------------------------------------------------------------------

	proc menubar_delete_scan_command { record_list } {
		set scan_record [ menubar_get_selected_scan $record_list ]

		set isa_scan [ $scan_record isa Mx::Scan ]

		if { $isa_scan == 0 } {
			set scan_name [ $scan_record get_field "name" ]

			set t "Record '$scan_name' is not a scan record\n"
			append t "so you cannot delete it."

			Mx::error_message "Cannot Delete Record" $t "" ""

			return
		}

		$scan_record delete_record

		$record_list configure -unsaved_scans 1
	}

	#------------------------------------------------------------------

	proc menubar_load_scans_dialog { record_list } {

		center_on_window . .scan_load_save_dialog

		.scan_load_save_dialog configure -selectionlabel \
			"Scan Database to Load"

		set result [ .scan_load_save_dialog activate ]

		if { $result == 0 } {
			# The user hit the Cancel button.

			return
		}

		set filename [ .scan_load_save_dialog get ]

		string trim $filename

		if { [string length $filename] == 0 } {

			Mx::error_message "Empty Filename" \
		"The scan database filename was empty.  No scans loaded." "" ""

			return
		}

		$record_list load_scans $filename 1

		.scan_load_save_dialog deactivate

		return
	}

	#------------------------------------------------------------------

	proc menubar_save_scans_dialog { record_list } {

		center_on_window . .scan_load_save_dialog

		.scan_load_save_dialog configure -selectionlabel \
			"Scan Database to Save To"

		set result [ .scan_load_save_dialog activate ]

		if { $result == 0 } {
			# The user hit the Cancel button.

			return
		}

		set filename [ .scan_load_save_dialog get ]

		string trim $filename

		if { [string length $filename] == 0 } {

			Mx::error_message "Empty Filename" \
		"The scan database filename was empty.  No scans loaded." "" ""

			return
		}

		$record_list save_scans $filename

		.scan_load_save_dialog deactivate

		return
	}

	#------------------------------------------------------------------

	proc menubar_run_scan_command { record_list } {
		set scan_record [ menubar_get_selected_scan $record_list ]

		#
		#  Check to see if the files created by this scan will
		#  overwrite already existing files.
		#

		set status [ mxgui_check_for_datafile_name_collision \
				$scan_record ]

		#
		#  If the returned status is 0, the user has decided to
		#  abort the scan since preexisting files would be
		#  overwritten.
		#

		if { $status == 0 } {
			return
		}

		#
		#  Prompt the user for scan header changes
		#

		set datafile_type [ get_datafile_type $scan_record ]

		set status [ menubar_setup_scan_header_change_dialog \
				$record_list $datafile_type $scan_record ]

		#
		#  If the returned status is 1, we have to do the scan ourself.
		#  Otherwise, a callback that was set up in the procedure
		#  menubar_setup_scan_header_change_dialog will do it for us.
		#

		if { $status == 1 } {
			mxgui_perform_scan $scan_record
		}
	}

	#------------------------------------------------------------------

	proc menubar_setup_scan_header_change_dialog \
				{ record_list datafile_type scan_record } {

		if { $datafile_type == "none" } {
			return 1
		}
		if { $datafile_type == "text" } {
			return 1
		}

		#
		#  Look for a variable of the form xxx_header1 where
		#  xxx is the datafile_type.  If there is no such variable,
		#  we don't have anything to change and can return.
		#

		set var_stem "${datafile_type}_header"

		set thisheader_name "${var_stem}1"

		set status [ catch { \
			$record_list get_record $thisheader_name } thisheader ]

		if { $status != 0 } {
			return 1
		}

		#
		#  Create a form that displays the header string values.
		#

		set toplevel .prompt_for_scan_header_change_dialog

		toplevel $toplevel

		wm title $toplevel "Scan File Header"

		frame $toplevel.title
		pack $toplevel.title

		label $toplevel.title.label -text "Scan File Header"
		pack $toplevel.title.label

		frame $toplevel.sep1 -height 2 -borderwidth 1 -relief sunken
		pack $toplevel.sep1 -expand yes -fill x

		set window $toplevel.main

		frame $window
		pack $window

		set i   1
		set row 0

		while { 1 } {

			set header_text [ $thisheader receive_variable ]

			set header_text [ string trim $header_text \" ]

			label $window.headername$i -text "$thisheader_name :"

			iwidgets::entryfield $window.header$i -width 80

			$window.header$i insert end $header_text

			grid $window.headername$i -row $row -column 0 \
							-sticky w -padx 5

			grid $window.header$i -row $row -column 1 -padx 5

			incr i
			incr row

			set thisheader_name "${var_stem}$i"

			set status [ catch { $record_list \
				get_record $thisheader_name } thisheader ]

			if { $status != 0 } {
				break
			}
		}

		set num_headers [ expr $i - 1 ]

		frame $toplevel.sep2 -height 2 -borderwidth 1 -relief sunken
		pack $toplevel.sep2 -pady 2 -expand yes -fill x

		iwidgets::buttonbox $toplevel.buttonbox
		pack $toplevel.buttonbox

		$toplevel.buttonbox add Continue -text Continue -command \
			[ itcl::code menubar_scan_header_change_continue \
				$record_list $datafile_type $num_headers \
				$scan_record $toplevel ]

		$toplevel.buttonbox add Cancel -text Cancel -command \
						[ itcl::code destroy $toplevel ]

		$toplevel.buttonbox default Continue

		return 0	;# Do not immediately run the scan.
	}

	#------------------------------------------------------------------

	proc menubar_scan_header_change_continue { record_list datafile_type \
					num_headers scan_record toplevel } {

		set window $toplevel.main

		for { set i 1 } { $i <= $num_headers } { incr i } {
			set header_text [ $window.header$i get ]

			# Put double quotes around the header text.

			set header_text "\"$header_text\""

			set header_name "${datafile_type}_header$i"

			set header_record \
				[ $record_list get_record $header_name ]

			$header_record send_variable $header_text
		}

		destroy $toplevel

		update

		mxgui_perform_scan $scan_record

		return
	}

	###################################################################
	#
	#  Add the Variable menu.
	#
	menu .menubar.variable -tearoff 0
	.menubar add cascade -label Variable -menu .menubar.variable

	.menubar.variable add command -label "Edit Variables" \
	    -command [ itcl::code menubar_edit_variables_command $record_list ]

	#------------------------------------------------------------------

	iwidgets::selectiondialog .variable_select_dialog

	.variable_select_dialog configure -title "Select variable"
	.variable_select_dialog configure -itemslabel Variables
	.variable_select_dialog configure -textfont $MxGui::monospaced_font
	.variable_select_dialog configure -height 350
	.variable_select_dialog configure -width 500

	.variable_select_dialog hide Apply
	.variable_select_dialog buttonconfigure OK -text "Select Variable"

	.variable_select_dialog buttonconfigure OK \
    -command [ itcl::code menubar_select_variable_ok_command $record_list ]

	proc menubar_edit_variables_command { record_list } {
		menubar_update_variable_list $record_list

		if { [ .variable_select_dialog size ] <= 0 } {

			Mx::error_message "No Variables Defined" \
	    "No variable records are defined for the current database." "" ""

		} else {
			center_on_window . .variable_select_dialog
			.variable_select_dialog activate
		}
	}

	proc menubar_update_variable_list { record_list } {
		set variable_list \
		    [ $record_list get_multiple_records "variable" ".*" ".*" 0 ]

		.variable_select_dialog clear items

		foreach name $variable_list {
			set variable_record [ $record_list get_record $name ]
			set variable_type [ $variable_record get_field mx_type ]
			set value_string [ $variable_record receive_variable ]

			set item_string [ format "%-24s  %-12s  %s" \
				$name $variable_type $value_string ]

			.variable_select_dialog insert items end $item_string
		}

		.variable_select_dialog selection set 0
	}

	proc menubar_select_variable_ok_command { record_list } {
		.variable_select_dialog deactivate

		set variable_item_string [ .variable_select_dialog get ]

		scan $variable_item_string "%s" variable_name

		set variable_record [ $record_list get_record $variable_name ]

		mxgui_create_variable_widget $variable_record
	}
	
	###################################################################
	#
	#  Add the Server menu.
	#
#	menu .menubar.server -tearoff 0
#	.menubar add cascade -label Server -menu .menubar.server -state disabled

	###################################################################
	#
	#  Add the Beamline menu.
	#
	menu .menubar.beamline -tearoff 0
	.menubar add cascade -label Beamline -menu .menubar.beamline

	# Get the name of the beamline specific menu entry script.

	set result \
	  [ catch {set mx_tcl_beamline_script $env(MX_TCL_BEAMLINE_SCRIPT)} ]

	if { $result != 0 } {
		set mx_tcl_beamline_script [ file join [ pwd ] "beamline.tcl" ]
	}

	# If the beamline specific script exists, then source it.

	if [ file isfile $mx_tcl_beamline_script ] {
		source $mx_tcl_beamline_script
	} else {
		.menubar entryconfigure Beamline -state disabled
	}

	###################################################################
	#
	#  Add the About menu.
	#
	menu .menubar.about -tearoff 0
	.menubar add cascade -label About -menu .menubar.about
	
	.menubar.about add command -label "About MX GUI" \
				-command menubar_about_command

	proc menubar_about_command {} {
		set t "MX GUI was written by William Lavender"
		tk_dialog .about_dialog "About MX GUI" $t info 0 "Done"
	}

}

#
#  Centers a window on the desktop.
#

proc center_on_screen { widget } {
	set xmax [ winfo screenwidth $widget ]
	set ymax [ winfo screenwidth $widget ]
	set x [ expr ($xmax - [ winfo reqwidth $widget ]) / 2 ]
	set y [ expr ($ymax - [ winfo reqheight $widget ]) / 2 ]
	wm geometry $widget "+$x+$y"
}

#
#  Centers a new window relative to an old window.
#

proc center_on_window { old_widget new_widget } {
	set old_width  [ winfo width  $old_widget ]
	set old_height [ winfo height $old_widget ]
	set new_width  [ winfo width  $new_widget ]
	set new_height [ winfo height $new_widget ]

	set old_ulcorner_x [ winfo rootx $old_widget ]
	set old_ulcorner_y [ winfo rooty $old_widget ]

	set xcenter [ expr ($old_ulcorner_x + 0.5 * $old_width) ]
	set ycenter [ expr ($old_ulcorner_y + 0.5 * $old_height) ]

	set new_ulcorner_x [ expr ($xcenter - 0.5 * $new_width) ]
	set new_ulcorner_y [ expr ($ycenter - 0.5 * $new_height) ]

	set new_ulcorner_x [ expr round($new_ulcorner_x) ]
	set new_ulcorner_y [ expr round($new_ulcorner_y) ]

	wm geometry $new_widget "+$new_ulcorner_x+$new_ulcorner_y"
}

